<?php

namespace Monarch;

/**
 * Provides color utilities for converting colors.
 */
abstract class AccessibleColor {

  const AA_TEXT_CONTRAST = 4.5;
  const AA_LARGE_TEXT_CONTRAST = 3.0;
  const AAA_TEXT_CONTRAST = 7.0;
  const AAA_LARGE_TEXT_CONTRAST = 4.5;
  const NON_TEXT_CONTRAST = 3.0;
  const LARGE_TEXT_BOLD_MIN_PX = 18.66;
  const LARGE_TEXT_BOLD_MIN_PT = 14;
  const LARGE_TEXT_NORMAL_MIN_PX = 24;
  const LARGE_TEXT_NORMAL_MIN_PT = 18;

  /**
   * Get recommended text contrast.
   *
   * @param float $text_size_px
   *   The text size in pixels.
   * @param bool $is_bold
   *   Is the text bold?
   * @param bool $is_aaa
   *   Do we need the more rigid AAA standards?
   *
   * @return float
   *   The recommended contrast ratio.
   */
  public static function getRecommendedContrast(float $text_size_px, bool $is_bold, bool $is_aaa = TRUE) : float {
    $text_size_threshold = $is_bold ? static::LARGE_TEXT_BOLD_MIN_PX : static::LARGE_TEXT_NORMAL_MIN_PX;

    [$large_text, $small_text] = $is_aaa ? [
      static::AAA_LARGE_TEXT_CONTRAST,
      static::AAA_TEXT_CONTRAST,
    ] : [
      static::AA_LARGE_TEXT_CONTRAST,
      static::AA_TEXT_CONTRAST,
    ];

    return $text_size_px >= $text_size_threshold ? $large_text : $small_text;
  }

  /**
   * Calculates the contrast ratio against another color.
   *
   * @param AccessibleColor $color
   *   The color to compare against.
   *
   * @return float
   *   The contrast ratio of the two colors.
   */
  public function getContrastRatio(AccessibleColor $color) : float {
    $c = [$this->getLuminance(), $color->getLuminance()];
    $c = [max(...$c), min(...$c)];
    return ($c[0] + 0.05) / ($c[1] + 0.05);
  }

  /**
   * Calculates the accessible version of a color.
   *
   * @param string $base_color
   *   The color to calculate against.
   * @param string $adj_color
   *   The color to adjust.
   * @param bool $is_text
   *   Is this text?
   * @param float $text_size_px
   *   The text size in pixels.
   * @param bool $is_bold
   *   Is the text bold?
   * @param bool $is_aaa
   *   Do we need the more rigid AAA standards?
   *
   * @return string
   *   The calculated color string.
   */
  public static function calcAccessibleColor(string $base_color, string $adj_color, bool $is_text = TRUE, float $text_size_px = 1.0, bool $is_bold = FALSE, bool $is_aaa = TRUE) : string {
    $contrast_ratio = $is_text ? static::getRecommendedContrast($text_size_px, $is_bold, $is_aaa) : static::NON_TEXT_CONTRAST;
    return RGB::fromHex($adj_color)->calcFromContrastRatio(RGB::fromHex($base_color), $contrast_ratio)->toHex();
  }

  /**
   * Calculate the color need to pass the contrast ratio minimum.
   *
   * @param AccessibleColor $base_color
   *   The color to compare against.
   * @param float $target_contrast_ratio
   *   The minimum contrast ratio to require.
   *
   * @return AccessibleColor
   *   The new color.
   */
  public function calcFromContrastRatio(AccessibleColor $base_color, float $target_contrast_ratio) : AccessibleColor {
    $base_color = $base_color->toRgb();
    $mid = $this->toHsl();
    $base_lum = $base_color->getLuminance();

    $dir = $mid->getLuminance() > $base_lum ? 1 : -1;

    for ($t = 0; $t < 2; $t++) {
      $upper = HSL::create(0, 0, $dir > 0 ? 1 : 0);
      $lower = $base_color;
      $mid->setLightness(($upper->getLightness() + $lower->getLightness()) / 2);

      $calc_ratio = function ($dir, $base_lum, $mid) {
        return $dir > 0 ?
          ($mid->getLuminance() + 0.05) / ($base_lum + 0.05) :
          ($base_lum + 0.05) / ($mid->getLuminance() + 0.05);
      };

      for ($c = 0; $c < 16; $c++) {
        if ($calc_ratio($dir, $base_lum, $mid) >= $target_contrast_ratio) {
          $upper = $mid->toHsl();
        }
        else {
          $lower = $mid->toHsl();
        }

        $mid->setLightness(($upper->getLightness() + $lower->getLightness()) / 2);
      }

      $mid = $mid->toHsl();

      $final_contrast = RGB::fromHex($mid->toHex())->getContrastRatio($base_color);
      while ($mid->getLightness() < 1 && $mid->getLightness() > 0 && $final_contrast < $target_contrast_ratio) {
        $mid->setLightness($mid->getLightness() + $dir / 256.0);
        $final_contrast = RGB::fromHex($mid->toHex())->getContrastRatio($base_color);
      }

      if ($final_contrast >= $target_contrast_ratio) {
        return $mid;
      }

      $dir *= -1;
    }

    throw new \Exception('Unable to find valid color!');
  }

  /**
   * Factory function.
   *
   * @param string $hex
   *   The red component.
   *
   * @return AccessibleColor
   *   The created color object.
   */
  abstract public static function fromHex(string $hex) : AccessibleColor;

  /**
   * Convert to string.
   *
   * @return string
   *   The string representation of the class.
   */
  abstract public function toHex() : string;

  /**
   * Convert to RGB.
   */
  abstract public function toRgb() : RGB;

  /**
   * Convert to HSL.
   */
  abstract public function toHsl() : HSL;

  /**
   * Calculates the luminance of a color.
   *
   * @return float
   *   The luminance of the color.
   */
  abstract public function getLuminance() : float;

  /**
   * Convert, clamp and return the lightness value.
   *
   * @return int
   *   The clamped and casted value.
   */
  abstract public function getLightness() : float;

  /**
   * Set the lightness of the color, converting if necessary.
   *
   * @param float $lightness
   *   The new lightness.
   *
   * @return AccessibleColor
   *   The current color object.
   */
  abstract public function setLightness(float $lightness) : AccessibleColor;

}
